from typing import Any

import requests
from requests import Response


class HttpResponse(object):
    response: Response = None
    event: Any = None

    def __init__(self, response: Response):
        self.response = response


class HttpClient(object):
    base_url: str = None
    secret_key: str = None
    headers: dict = {}

    def __init__(self, secret_key: str, base_url: str = None):
        self.secret_key = secret_key
        self.base_url = base_url or "https://api.waitlyst.co/v1"
        headers = {
            'Content-Type': 'application/json',
            'Authorization': f"Bearer {self.secret_key}"
        }
        self.set_headers(headers)

    def set_headers(self, headers: dict):
        """Set the headers for the HTTP client."""
        self.headers = headers

    def post(self, path: str, data: dict, headers: dict = None) -> Response:
        """Send a POST request to the API."""
        if headers is None:
            headers = self.headers
        url = f"{self.base_url}{path}"
        response = requests.post(url=url, json=data, headers=headers)
        return response

    def get(self, path: str, params: dict, headers: dict = None) -> Response:
        """Send a GET request to the API."""
        if headers is None:
            headers = self.headers
        url = f"{self.base_url}{path}"
        response = requests.get(url, params=params, headers=headers)
        return response

    def put(self, path: str, data: dict, headers: dict = None) -> Response:
        """Send a PUT request to the API."""
        if headers is None:
            headers = self.headers
        url = f"{self.base_url}{path}"
        response = requests.put(url, json=data, headers=headers)
        return response

    def patch(self, path: str, data: dict, headers: dict = None) -> Response:
        """Send a PATCH request to the API."""
        if headers is None:
            headers = self.headers
        url = f"{self.base_url}{path}"
        response = requests.patch(url, json=data, headers=headers)
        return response

    def delete(self, path: str, headers: dict = None) -> Response:
        """Send a DELETE request to the API."""
        if headers is None:
            headers = self.headers
        url = f"{self.base_url}{path}"
        response = requests.delete(url, headers=headers)
        return response



EVENT_TEMPLATE = {
    "anonymousId": "",
    "messageId": "",
    "receivedAt": "",
    "sentAt": "",
    "timestamp": "",
    "type": "",
    "userId": "",
    "version": 1,
}

API_PATHS = {
    "process_event": "/events/process/",
}



class BaseException(Exception):
    """Base class for all exceptions."""

    def __init__(self, message):
        self.message = message

    def __repr__(self):
        return self.message


class ErrorSecretKeyRequired(BaseException):
    message = "You must set a secret key to use waitlyst."


class ErrorEventTypeRequired(Exception):
    message = "You must set an event type."


class ErrorEventNameRequired(Exception):
    message = "You must set an name for this event."


class ErrorInvalidPayload(Exception):
    message = "The payload is invalid."



import uuid
from datetime import datetime
from typing import Optional, List

from requests import Response

from waitlyst.constants import API_PATHS, EVENT_TEMPLATE
from waitlyst.exceptions import (
    ErrorEventTypeRequired,
    ErrorEventNameRequired,
    ErrorInvalidPayload,
)
from waitlyst.http import HttpClient, HttpResponse
from waitlyst.models import User


class Event(object):
    type: str = None
    user_id: Optional[str] = None
    anonymous_id: Optional[str] = None
    timestamp: Optional[str] = None
    context: Optional[dict] = None
    message_id: Optional[str] = None
    received_at: Optional[str] = None
    sent_at: Optional[str] = None
    status: Optional[str] = None

    def __init__(self, **kwargs):
        self.type = kwargs.get("type")
        self.user_id = kwargs.get("userId")
        self.anonymous_id = kwargs.get("anonymousId")
        self.timestamp = kwargs.get("timestamp")
        self.context = kwargs.get("context")
        self.message_id = str(uuid.uuid4())
        self.received_at = datetime.now().isoformat()
        self.sent_at = datetime.now().isoformat()


class TrackEvent(Event):
    type = "track"
    event: str = None
    properties: Optional[dict] = None


class PageEvent(Event):
    type = "page"
    name: str = None
    properties: Optional[dict] = None


class IdentifyEvent(Event):
    type = "identify"
    traits: Optional[dict] = None


class EventManager(object):
    secret_key: str = None
    user: User = None
    client: HttpClient = None
    queue: List[Event] = []

    def __init__(self, secret_key: str, base_uri: str = None):
        self.secret_key = secret_key
        self.user = User()
        self.client = HttpClient(secret_key=secret_key, base_url=base_uri)

    def set_anonymous_id(self, anonymous_id: str) -> User:
        self.user.anonymous_id = anonymous_id
        return self.user

    def reset(self):
        self.user = User()

    def identity(self) -> User:
        return self.user

    def construct(
        self, type: str, name: Optional[str] = None, props: Optional[dict] = None
    ) -> dict:
        """Constructs a payload in the appropriate format."""
        if not type:
            raise ErrorEventTypeRequired
        if not name and type != "identify":
            raise ErrorEventNameRequired
        now = datetime.now().isoformat("#", "milliseconds")
        payload = EVENT_TEMPLATE.copy()
        payload["anonymousId"] = self.user.anonymous_id
        payload["messageId"] = str(uuid.uuid4())
        payload["receivedAt"] = now
        payload["sentAt"] = now
        payload["timestamp"] = now
        payload["type"] = type
        payload["userId"] = self.user.id

        if type == "track":
            payload["event"] = name
            payload["properties"] = props
        elif type == "page":
            payload["name"] = props
            payload["properties"] = props
        elif type == "identify":
            payload["traits"] = props
            payload["userId"] = name

        return payload

    def create_event(self, payload):
        """Create an event object from a payload"""
        if payload["type"] == "track":
            return TrackEvent(**payload)
        elif payload["type"] == "page":
            return PageEvent(**payload)
        elif payload["type"] == "identify":
            return IdentifyEvent(**payload)
        else:
            raise ErrorInvalidPayload

    def handle(self, response: Response, payload: dict) -> HttpResponse:
        """Handle the response from the API."""
        event = self.create_event(payload)
        event.status = "success" if response.status_code == 200 else "error"
        http_response = HttpResponse(response)
        http_response.event = event

        # Push to queue
        self.queue.append(event)

        if response.status_code == 200 and payload["type"] == "identify":
            self.user.id = payload["userId"]

        return http_response

    def identify(self, id: str = None, traits: Optional[dict] = None) -> HttpResponse:
        payload = self.construct("identify", id, traits)
        response = self.client.post(path=API_PATHS["process_event"], data=payload)
        return self.handle(response, payload)

    def track(self, event: str, properties: Optional[dict] = None) -> HttpResponse:
        payload = self.construct("track", event, properties)
        response = self.client.post(path=API_PATHS["process_event"], data=payload)
        return self.handle(response, payload)

    def page(self, name: str, properties: Optional[dict] = None) -> HttpResponse:
        payload = self.construct("page", name, properties)
        response = self.client.post(path=API_PATHS["process_event"], data=payload)
        return self.handle(response, payload)



from typing import Optional, List

from waitlyst.events import EventManager, Event
from waitlyst.exceptions import ErrorSecretKeyRequired
from waitlyst.http import HttpResponse
from waitlyst.models import User


class Waitlyst(object):
    secret_key: Optional[str] = None
    event_manager: Optional[EventManager] = None

    def __init__(self, secret_key: str):
        self.secret_key = secret_key
        self.event_manager = EventManager(self.secret_key)
        self.initialize()

    def initialize(self):
        """Run through checks ensuring all necessary values are filled."""
        if not self.secret_key:
            raise ErrorSecretKeyRequired

    def identify(self, id: str = None, traits: Optional[dict] = None) -> HttpResponse:
        """Identify a user."""
        return self.event_manager.identify(id, traits)

    def set_anonymous_id(self, anonymous_id: str) -> User:
        """Set the anonymous id."""
        return self.event_manager.set_anonymous_id(anonymous_id)

    def track(self, event: str, properties: Optional[dict] = None) -> HttpResponse:
        """Track an event."""
        return self.event_manager.track(event, properties)

    def page(self, name: str, properties: Optional[dict] = None) -> HttpResponse:
        """Track a page view."""
        return self.event_manager.page(name, properties)

    def identity(self) -> User:
        """Get the current user."""
        return self.event_manager.identity()

    @property
    def queue(self) -> List[Event]:
        """Get the current queue."""
        return self.event_manager.queue

    def reset(self):
        """Resets the current context."""
        self.event_manager.reset()
        self.event_manager.queue = []



from typing import Optional


class User(object):
    id: Optional[str] = None
    anonymous_id: Optional[str] = None

    def __init__(self, id: str = None, anonymous_id: str = None):
        self.id = id
        self.anonymous_id = anonymous_id

    def __str__(self):
        return f"User(id={self.id}, anonymous_id={self.anonymous_id})"

    def __repr__(self):
        return self.__str__()

    def reset(self):
        """Reset the user to an anonymous user."""
        self.id = None
        self.anonymous_id = None
